/*
- extremeley wide images ae skipped at the moment

*/

#include "SDLTest.h"

int global_message_output;

#define MSG_TXT_USAGE               0
#define MSG_TXT_WRONG_THUMB_QUALITY 1
#define MSG_TXT_WRONG_THUMB_HEIGHT  2


void MessageText(int id_text, bool debug)
{
    char* p_message;

    switch(id_text) {
	case MSG_TXT_USAGE:
	    p_message = "usage: picturespace [options] path\n\n"\
		"'path' should point to a directory containing images (jpeg, bmp)\n"\
		"or containing subdirectories with images if combined with '-r'\n\n"\
		"options:\n"\
		"-r          recurse into subdirectories below 'path'\n"\
		"-q n        thumbnail quality (n=0: low/fast (default), ..., n=5: hight/slow)\n"\
		"-t n        height of thumbnail in pixels (n=12, ..., 255)\n"\
		"-m n        message output (n=0: quiet / n=1: verbos / n=2: debug)\n"\
		"-d n        n=0: quiet / n=1: verbos / n=2: debug (quite a lot)\n"\
		"-s n        start width resolution n=0: 640x480 / n=1: 800x600 / n=2: 1024x768\n"\
		"                                   n=3: 1280x1024 / n=4: 1600x1200\n"\
		"-h          display this help text\n\n"\
		"example: picturespace -r ~\n";
	    break;

	case MSG_TXT_WRONG_THUMB_QUALITY:
	    p_message = "thumbnail quality should be within 0, ..., 5\n"\
	                "0 = low image quality/ high speed ...\n"\
		"5 = highest image quality/ low speed\n\n";
	    break;

	case MSG_TXT_WRONG_THUMB_HEIGHT:
	    p_message = "thumbnail height should be within 12, ..., 255\n\n";
	    break;

	default:
	    p_message = "unknown message id in MessageText()";
	    break;
    } // switch

    if(debug) cerr << endl << p_message;
    else cout << endl << p_message;
} // MessageText

Image::Image(char* p_filename, char* p_path, long x, SDL_Surface* p_image, SDL_Surface* p_thumb)
{
    _p_filename = p_filename;
    _p_path = p_path;

    _location_x = x;

    _p_image = p_image;
    _p_thumb = p_thumb;
} // Image::Image

Image::~Image()
{
    if(_p_thumb) SDL_FreeSurface(_p_thumb);
    if(_p_image) SDL_FreeSurface(_p_image);
} // Image::~Image

void Image::DisplayImage(SDL_Surface* dest_surface, SDL_Rect* p_dst_rec)
{
    SDL_Surface* p_display_image;
    bool         selfmade = false;
    SDL_Rect     src_rec;
    SDL_Rect     dst_rec;

    dst_rec.x = p_dst_rec->x;
    dst_rec.y = p_dst_rec->y;
    dst_rec.w = p_dst_rec->w;
    dst_rec.h = p_dst_rec->h;

    p_display_image = GetImage();

    if((dst_rec.w > GetImage()->w) || (dst_rec.h > GetImage()->h)) {
	p_display_image = ResizeImage(GetImage(), dst_rec.h, 0);
	selfmade = true;
    } // if

    if((dst_rec.w < GetImage()->w) || (dst_rec.h < GetImage()->h)) {
	p_display_image = ResizeImage(GetImage(), dst_rec.h, 0);
	selfmade = true;
    } // if

    src_rec.x = 0; src_rec.y = 0;
    src_rec.w = p_display_image->w; src_rec.h = p_display_image->h; 

//    SDL_SetAlpha(p_display_image, SDL_SRCALPHA, 128);

    for(int y = p_dst_rec->y; y < p_dst_rec->y + p_dst_rec->h; y += (p_dst_rec->h / SMOOTH_REBUILD_UNITS)) {

	dst_rec.x = p_dst_rec->x;
	dst_rec.w = p_dst_rec->w;
	dst_rec.h = p_dst_rec->h / SMOOTH_REBUILD_UNITS;
	dst_rec.y = y;

//	SDL_SetAlpha(p_display_image, SDL_SRCALPHA, (y - p_dst_rec->y)*255/p_dst_rec->h);

	if((dst_rec.y + dst_rec.h) > (p_dst_rec->y + p_dst_rec->h)) dst_rec.h = dst_rec.y + p_dst_rec->h - p_dst_rec->y;
	
	src_rec.x = 0; src_rec.y = y - p_dst_rec->y;
	src_rec.w = dst_rec.w; src_rec.h = dst_rec.h; 

	SDL_BlitSurface(p_display_image, &src_rec, dest_surface, &dst_rec);
	SDL_UpdateRect(dest_surface, dst_rec.x, dst_rec.y, dst_rec.w, dst_rec.h);

	usleep(1);
    } // for
/*
    dst_rec.x = p_dst_rec->x;
    dst_rec.w = p_dst_rec->w;
    dst_rec.h = p_dst_rec->h;
    dst_rec.y = p_dst_rec->y;

    SDL_SetAlpha(p_display_image, 0, 255);

    SDL_BlitSurface(p_display_image, NULL, dest_surface, &dst_rec);
    SDL_UpdateRect(dest_surface, dst_rec.x, dst_rec.y, dst_rec.w, dst_rec.h);
*/
    if(selfmade) SDL_FreeSurface(p_display_image);
} // Image::DisplayImage

SDL_Surface* Image::GetImage()
{
    if(!_p_image) {
	if(global_message_output >= MESSAGE_OUTPUT_DEBUG)
	    cout << "image " << _p_filename << " was not found in cache so load" << endl;

	_p_image = IMG_Load(_p_filename);
    } // if

    return _p_image;
} // Image::GetImage

void Image::WriteToCache(FILE* p_index, FILE* p_cache)
{
    // write filename incl. path into index_file

    fwrite(_p_filename, strlen(_p_filename)+1, 1, p_index);

    // write thumbnail into cache_file

    short w, h;

    w = _p_thumb->w;
    fwrite(&w, sizeof(w), 1, p_cache);

    h = _p_thumb->h;
    fwrite(&h, sizeof(h), 1, p_cache);

    fwrite(_p_thumb->pixels, w*h*3, 1, p_cache);
} // Image::WriteToCache

ThumbLine::ThumbLine()
{
    _p_image_array = NULL;
    _max_images = _index_next_image = 0;
} // ThumbLine::ThumbLine

ThumbLine::~ThumbLine()
{
    for(int i = 0; i < _index_next_image; i++) {
	delete _p_image_array[i];
    } // for

    if(_p_image_array) delete[] _p_image_array;
} // ThumbLine::~ThumbLine

void ThumbLine::AddImage(Image* p_image)
{
    Image** p_temp_image_array;

    if(_index_next_image >= _max_images) {
	_max_images += GROW_IMAGES_PER_LINE;
	p_temp_image_array = new Image* [_max_images];

	bcopy((void*)_p_image_array, (void*)p_temp_image_array, sizeof(Image*)*_index_next_image);
	delete[] _p_image_array;
	_p_image_array = p_temp_image_array;
    } // if

    _p_image_array[_index_next_image] = p_image;
    _index_next_image++;
} // ThumbLine::AddImage

void ThumbLine::BlockShift(long window_width)
{
    long total_width = 0;

    if(_index_next_image < 2) return;

    // add cummulated width of all image blocks

    for(int i = 0; i < _index_next_image; i++) {
	total_width += _p_image_array[i]->GetThumb()->w;
    } // for

    long space_avail = window_width - total_width;
    long average_space = space_avail / _index_next_image;
    long current_width = 0;

    // reposition images for optimal fit

    for(int i = 0; i < _index_next_image; i++) {
	_p_image_array[i]->SetXStart(current_width + space_avail*i/(_index_next_image-1));
	current_width += _p_image_array[i]->GetThumb()->w;
    } // for
} // ThumbLine::BlockShift

Image* ThumbLine::GetImage(long x)
{
    int      i;
    SDL_Rect src_rec;
    
    for(i = 0; i < _index_next_image; i++) {
	if((x >= _p_image_array[i]->GetXStart()) && (x < _p_image_array[i]->GetXEnd())) break;
    } // for

    if(i == _index_next_image) return NULL;

 display_image:

    return _p_image_array[i];
} // ThumbLine::GetImage

void ThumbLine::Display(SDL_Surface* p_screen, long y, long thumb_height)
{
    SDL_Rect rect;

    for(int i = 0; i < _index_next_image; i++) {
	rect.x = _p_image_array[i]->GetXStart(); rect.y = y;
	rect.w = _p_image_array[i]->GetThumb()->w; rect.h = thumb_height;

	SDL_BlitSurface(_p_image_array[i]->GetThumb(), NULL, p_screen, &rect);
	SDL_UpdateRect(p_screen, rect.x, y, _p_image_array[i]->GetThumb()->w, thumb_height);

    } // for
} // ThumbLine::Display

void ThumbLine::Save(FILE* p_index, FILE* p_cache)
{
    for(int i = 0; i < _index_next_image; i++) {
	_p_image_array[i]->WriteToCache(p_index, p_cache);
    } // for    
} // ThumbLine::Save

char* GetExtension(char* p_filename)
{
    char* p_work_text;

    p_work_text = p_filename + strlen(p_filename) - 1;

    while((*p_work_text != '.') && (p_work_text != p_filename))
	p_work_text--;

    if(p_work_text == p_filename) return NULL;

    return p_work_text+1;
} // GetExtension

void ScanImageFiles(char* p_start_path, char*& p_filenames, char*& p_next_file, long &max_filenames, bool recursive)
{
    char           *p_temp_filenames;
    char*          p_extension, text[512], current_path[512];
    long           current_buffer_size;
    DIR*           p_dir;
    struct dirent* p_next_entry;
    struct stat file_stat;

    p_dir = opendir(p_start_path);
   	    
    if(p_dir) {

	if((global_message_output > MESSAGE_OUTPUT_QUIET) && (global_message_output < MESSAGE_OUTPUT_DEBUG))
	    cout << "." << flush; // print one dot for each directory

	while((p_next_entry = readdir(p_dir)) != NULL) {

	    if((strcmp(p_next_entry->d_name, ".") != 0) && (strcmp(p_next_entry->d_name, "..") != 0)) {

		sprintf(current_path, "%s/%s", p_start_path, p_next_entry->d_name);
		stat(current_path, &file_stat);

		if((recursive) && ((file_stat.st_mode & S_IFDIR) != 0)) {

		    if(global_message_output >= MESSAGE_OUTPUT_DEBUG)      
			cout << "recurse into directory: " << current_path << endl;

		    ScanImageFiles(current_path, p_filenames, p_next_file, max_filenames, recursive);
		} // if
		else {
		    
		    p_extension = GetExtension(p_next_entry->d_name);
		    
		    if(p_extension) {
			if((strcasecmp(p_extension, "jpg") == 0) || (strcasecmp(p_extension, "jpeg") == 0) || (strcasecmp(p_extension, "_bmp_") == 0)) {
			    
			    // maybe the filename buffer has to grow
			    
			    current_buffer_size = (p_next_file - p_filenames);

			    sprintf(text, "%s/%s", p_start_path, p_next_entry->d_name);
			    
			    if((current_buffer_size + strlen(text)+5) >= max_filenames) {
				if(global_message_output >= MESSAGE_OUTPUT_DEBUG)      
				    cout << "filebuffer overflow. try the extend by (bytes): " << FILE_BUFFER_GROW_SIZE << endl;


				max_filenames += FILE_BUFFER_GROW_SIZE;
				p_temp_filenames = new char [max_filenames];
				
				if(p_filenames) {
				    bcopy(p_filenames, p_temp_filenames, current_buffer_size);
				    delete[] p_filenames;
				} // if
				
				p_filenames = p_temp_filenames;
				p_next_file = p_filenames + current_buffer_size;
			    } // if

			    if(global_message_output >= MESSAGE_OUTPUT_DEBUG)
				cout << "add new image file: " << text << endl;

			    bcopy(text, p_next_file, strlen(text));
			    p_next_file += strlen(text);
			    *p_next_file = 0; p_next_file++;
			} // if
		    } // if
		} // else
	    } // if
	} // while

	closedir(p_dir);
	if(p_next_file) *p_next_file = 0; // put and end to the string
    } // if
} // ScanImageFiles

SDL_Surface* ResizeImage(SDL_Surface* p_image, long height, short thumb_quality)
{
    SDL_Surface   *p_new_image = NULL;
    long          inner_count, outer_count, width, step_size;
    long          sum_red, sum_green, sum_blue, small_x, small_y;
    long          *p_tmp_red, *p_tmp_green, *p_tmp_blue;
    unsigned char *p_current_pixel, *p_work_orig_pixels, *p_small_pixels;

    width = p_image->w*height / p_image->h;
    p_new_image = SDL_CreateRGBSurface(SDL_SWSURFACE, width, height, 24, p_image->format->Rmask, p_image->format->Gmask, p_image->format->Bmask, p_image->format->Amask);

    if(height < p_image->h) {

	p_tmp_red = new long [width];
	p_tmp_green = new long [width];
	p_tmp_blue = new long [width];

	// the following procedure works
	// only for images being shrinked

	bzero(p_tmp_red, width*sizeof(long));
	bzero(p_tmp_green, width*sizeof(long));
	bzero(p_tmp_blue, width*sizeof(long));

	outer_count = 0; small_y = 0;
	step_size = 6-thumb_quality;

	if(step_size > (p_image->w/width)) step_size = (p_image->w/width);
	if(step_size > (p_image->h/height)) step_size = (p_image->h/height);

	for(long y = 0; y < p_image->h; y += step_size) {

	    // reset the current line

	    unsigned char* p_orig_pixels = &((unsigned char*)p_image->pixels)[y*p_image->pitch];

	    inner_count = 0; small_x = 0;
	    sum_red = sum_green = sum_blue = 0;

	    for(long x = 0; x < p_image->w; x += step_size) {

		sum_red += (long)p_orig_pixels[x*3+0];
		sum_green += (long)p_orig_pixels[x*3+1];
		sum_blue += (long)p_orig_pixels[x*3+2];
		inner_count++;

		if(x*width/p_image->w > small_x) {

		    if(small_x < p_new_image->w) {
		    
			// let's advance one pixel in the small image

			p_tmp_red[small_x] += (sum_red / inner_count);
			p_tmp_green[small_x] += (sum_green / inner_count);
			p_tmp_blue[small_x] += (sum_blue / inner_count);

			small_x++;
		    } // if

		    inner_count = 0;
		    sum_red = sum_green = sum_blue = 0;
		} // if
	    } // for

	    outer_count++;

	    if(y*height/p_image->h > small_y) {
		
		// let's advance one line in the small image

		if(small_y < p_new_image->h) {

		    p_small_pixels = &((unsigned char*)p_new_image->pixels)[small_y*p_new_image->pitch];

		    for(long x = 0; x < width; x++) {
			p_small_pixels[x*3+0] = p_tmp_red[x] / outer_count;
			p_small_pixels[x*3+1] = p_tmp_green[x] / outer_count;
			p_small_pixels[x*3+2] = p_tmp_blue[x] / outer_count;
		    
		    } // for

		    bzero(p_tmp_red, width*sizeof(long));
		    bzero(p_tmp_green, width*sizeof(long));
		    bzero(p_tmp_blue, width*sizeof(long));

		    outer_count = 0;
		    small_y++;
		} // if
	    } // if
	} // for
	
	delete[] p_tmp_blue;
	delete[] p_tmp_green;
	delete[] p_tmp_red;
    } // if
    else {
	
	// this procedure works only
	// for images being enlarged

	for(long y = 0; y < height; y++) {
	    for(long x = 0; x < width; x++) {
		    (((unsigned char*)p_new_image->pixels)[x*3 + y*p_new_image->pitch + 0]) = (((unsigned char*)p_image->pixels)[(x*p_image->w/width)*3 + (y*p_image->h/height)*p_image->pitch + 0]);
		    (((unsigned char*)p_new_image->pixels)[x*3 + y*p_new_image->pitch + 1]) = (((unsigned char*)p_image->pixels)[(x*p_image->w/width)*3 + (y*p_image->h/height)*p_image->pitch + 1]);
		    (((unsigned char*)p_new_image->pixels)[x*3 + y*p_new_image->pitch + 2]) = (((unsigned char*)p_image->pixels)[(x*p_image->w/width)*3 + (y*p_image->h/height)*p_image->pitch + 2]);
	    } // for
	} // for

    } // else

    return p_new_image;
} // ResizeImage

/*

SDL_Surface* ResizeImage(SDL_Surface* p_image, long height, short thumb_quality)
{
    SDL_Surface   *p_small_image;
    long          count, width, red, green, blue;
    unsigned char *p_current_pixel, *p_work_orig_pixels;
    double        aspect_ratio_x, aspect_ratio_y;

    width = p_image->w*height / p_image->h;
    p_small_image = SDL_CreateRGBSurface(SDL_SWSURFACE, width, height, 24, p_image->format->Rmask, p_image->format->Gmask, p_image->format->Bmask, p_image->format->Amask);

    aspect_ratio_x = (double)p_image->w/(double)width;
    aspect_ratio_y = (double)p_image->h/(double)height;

    for(long y = 0; y < height; y++) {

	p_current_pixel =  &((unsigned char*)p_small_image->pixels)[y*p_small_image->pitch];

	for(long x = 0; x < width; x++) {

	    count = 0;
	    red = green = blue = 0;

	    for(double dy = 0; dy <= ((double)thumb_quality/5.0); dy += 1.0/aspect_ratio_y) {
		p_work_orig_pixels = &(((unsigned char*)p_image->pixels)[(long)((double)x*aspect_ratio_x)*3 + ((int)(((double)y+dy)*aspect_ratio_y)*p_image->pitch)+0]);

		for(long scan_x = (long)((double)x*aspect_ratio_x); scan_x <= (long)(((double)x+(double)thumb_quality/5.0)*aspect_ratio_x); scan_x++) {

		    red += (long)*p_work_orig_pixels; p_work_orig_pixels++;
		    green += (long)*p_work_orig_pixels; p_work_orig_pixels++;
		    blue += (long)*p_work_orig_pixels; p_work_orig_pixels++;
		    count++;

		} // for
	    } // for

	    *p_current_pixel = red / count; p_current_pixel++;
	    *p_current_pixel = green / count; p_current_pixel++;
	    *p_current_pixel = blue / count; p_current_pixel++;
	} // for
    } // for

    return p_small_image;
} // ResizeImage

*/

bool CheckFileIntegrity(char* p_filename)
{
    struct stat file_stat;

    stat(p_filename, &file_stat);
    if(file_stat.st_size < 120) return false; // can not be a valid jpeg file

    return true;
} // CheckFileIntegrity

ThumbLine* LoadNextThumbLine(char*& p_current_filename, long window_width, short thumb_quality, long thumb_height, long thumb_border_size, char* p_image_path = NULL)
{
    SDL_Surface* p_jpeg_image, *p_thumb_image;
    ThumbLine*   p_thumb_line = new ThumbLine();
    char         full_filename[512], *p_next_filename;
    long         delta_x = 0;

    p_next_filename = p_current_filename;

    while(*p_next_filename != 0) {

	if(p_image_path) sprintf(full_filename, "%s/%s", p_image_path, p_next_filename);
	else  sprintf(full_filename, "%s", p_next_filename);

	if(global_message_output >= MESSAGE_OUTPUT_DEBUG)
	    cout << "trying to load " << full_filename << " ... " << flush;

	p_jpeg_image = NULL;

	if(CheckFileIntegrity(full_filename)) {
	    p_jpeg_image = IMG_Load(full_filename);
	} // if
	else {
	    if(global_message_output >= MESSAGE_OUTPUT_DEBUG)
		cout << "not a vaild jpeg file" << endl << flush;
	} // else
  
	if(global_message_output >= MESSAGE_OUTPUT_DEBUG) {
	    if(p_jpeg_image) cout << "ok" << endl << flush;
	    else cout << "error" << endl << flush;
	} // if

	// don't include images smaller than thumb_height
/*
	if((p_jpeg_image) && (p_jpeg_image->h < thumb_height)) {
	    SDL_FreeSurface(p_jpeg_image);
	    p_jpeg_image = NULL;
	} // if
*/
	if(p_jpeg_image) {
	    SDL_Surface* p_thumb_image = ResizeImage(p_jpeg_image, thumb_height, thumb_quality);

	    SDL_FreeSurface(p_jpeg_image); p_jpeg_image = NULL;

	    if((delta_x + p_thumb_image->w) >= window_width) {

		short temp_width = p_thumb_image->w;

		SDL_FreeSurface(p_thumb_image);

		if(temp_width >= window_width)
	       	    goto try_next_image;
		else break;
	    } // if
	    
	    Image* p_new_image = new Image(p_next_filename, p_image_path, delta_x, p_jpeg_image, p_thumb_image);
	    
	    p_thumb_line->AddImage(p_new_image);
	    delta_x += p_thumb_image->w + thumb_border_size;
	} // if

    try_next_image:
	p_next_filename += strlen(p_next_filename) + 1;
	p_current_filename = p_next_filename;
    } // while

    if(delta_x == 0) {
	delete p_thumb_line;
	p_thumb_line = NULL;
    } // if
    else p_thumb_line->BlockShift(window_width);

    return p_thumb_line;
} // LoadNextThumbLine

void BackupImage(SDL_Surface* p_screen, SDL_Surface* p_backup_screen)
{
    SDL_BlitSurface(p_backup_screen, NULL, p_screen, NULL);
} // BackupImage

void RestoreImage(SDL_Surface* p_screen, SDL_Surface* p_backup_screen, SDL_Rect   location, bool *p_image_open)
{
    SDL_Rect  rect;

    if(*p_image_open) {

	for(int y = location.y + location.h; y >= location.y; y -= (location.h / SMOOTH_REBUILD_UNITS)) {

	    rect.x = location.x;
	    rect.w = location.w;
	    rect.h = location.h / SMOOTH_REBUILD_UNITS;
	    rect.y = y - rect.h;

	    SDL_BlitSurface(p_backup_screen, &rect, p_screen, &rect);
	    SDL_UpdateRect(p_screen, rect.x, rect.y, rect.w, rect.h);

	    usleep(1);
	} // for

	*p_image_open = false;
    } // if
} // RestoreImage

int main(int argc, char** argv)
{
    char*            p_next_filename, clicked_thumb_line;
    char*            p_image_path, *p_filenames;
    SDL_Surface      *screen, *backup_screen;
    SDL_Event        event;
    SDL_Rect         dst_rect, full_image_rect, scrolled_rect, location;
    unsigned char*   p_pixel;
    bool             first_time, image_open;
    ThumbLine*       p_thumb_line_array[MAX_THUMB_LINES];
    Image*           p_big_image;
    long             index_thumb_line, offset_display_index;
    long             no_displayable_lines, max_filenames, delta_x, delta_y;
    bool             descend_into_directories;
    long             thumb_height, thumb_border_size, window_width, window_height;
    int              c, option_index;
    int              thumb_quality = 0;

    window_width = 1280; window_height = 1024;
    thumb_height = 128; thumb_border_size = 12;

    global_message_output = MESSAGE_OUTPUT_NORMAL;
    p_image_path = NULL;

    descend_into_directories = false;

    // analyze command line

    while(true) {

	c = getopt(argc, argv, "hm:rq:t:s:");

	if(c == -1) break;

	switch(c) {
	    case 'h':
		MessageText(MSG_TXT_USAGE, false);
		exit(0);
		break;

	    case 'm':
		switch(atoi(optarg)) {
		    case 0:
			global_message_output = MESSAGE_OUTPUT_QUIET;
			break;

		    case 1:
			global_message_output = MESSAGE_OUTPUT_VERBOSE;
			break;

		    case 2:
			global_message_output = MESSAGE_OUTPUT_DEBUG;
			break;

		} // switch
		break;
	    
	    case 'r':
		descend_into_directories = true;
		break;

	    case 'q':
 		if(optarg) thumb_quality = atoi(optarg);
		if((thumb_quality < 0) || (thumb_quality > 5)) {
		    MessageText(MSG_TXT_WRONG_THUMB_QUALITY, true);
		    exit(-1);
		} // if
		break;

	    case 't':
 		if(optarg) thumb_height = atoi(optarg);
		if((thumb_height < 12) || (thumb_height > 255)) {
		    MessageText(MSG_TXT_WRONG_THUMB_HEIGHT, true);
		    exit(-1);
		} // if
		break;

	    case 's':
		switch(atoi(optarg)) {
		    case 0:
			window_width = 640;
			window_height = 480;
			break;

		    case 1:
			window_width = 800;
			window_height = 600;
			break;

		    case 2:
			window_width = 1024;
			window_height = 768;
			break;

		    case 3:
			window_width = 1280;
			window_height = 1024;
			break;

		    case 4:
			window_width = 1600;
			window_height = 1200;
			break;

		} // switch
		break;

	    default:
		cerr << "no match: " << c << endl;
		break;
	} // switch
    } // while

    if(optind < argc) 
	p_image_path = argv[optind];
    else {
	cerr << argv[0] << ": path argument is missing - abort" << endl;
	MessageText(MSG_TXT_USAGE, false);
	exit(-1);
    } // else

    // init

    p_next_filename = p_filenames = NULL;
    max_filenames = 0;

    for(int i = 0; i < MAX_THUMB_LINES; i++)
	p_thumb_line_array[i] = NULL;

    no_displayable_lines = window_height / (thumb_height + thumb_border_size);

    index_thumb_line = 0;
    offset_display_index = 0;

    if(global_message_output > MESSAGE_OUTPUT_QUIET)
	cout << "scanning for images";
		
    ScanImageFiles(p_image_path, p_filenames, p_next_filename, max_filenames, descend_into_directories);
//    *p_next_filename = 0;

    if(global_message_output > MESSAGE_OUTPUT_QUIET)
	cout << endl;

    if(!p_filenames) {
	cerr << argv[0] << ": no image files found - exit" << endl;
	exit(0);
    } // if

    if(SDL_Init(SDL_INIT_VIDEO) < 0) {
	cerr << "Could not initialize SDL video" << endl;
	exit(-1);
    } // if
    atexit(SDL_Quit);

    screen = SDL_SetVideoMode(window_width, window_height, 24, SDL_FULLSCREEN);
//    screen = SDL_SetVideoMode(window_width, window_height, 24, SDL_SWSURFACE);
    backup_screen = SDL_CreateRGBSurface(SDL_SWSURFACE, window_width, window_height, 24, screen->format->Rmask, screen->format->Gmask, screen->format->Bmask, screen->format->Amask);

    if(!screen) {
	cerr << argv[0] << "could not SetVideoMode() - abort" << endl;
	exit(-1);
    } // if

    delta_x = 0; delta_y = 0;
    p_next_filename = p_filenames;

    while((index_thumb_line+1)*(thumb_height + thumb_border_size) < window_height) {

	if((p_thumb_line_array[index_thumb_line] = LoadNextThumbLine(p_next_filename, window_width, thumb_quality, thumb_height, thumb_border_size)) == NULL) break;
	    
	p_thumb_line_array[index_thumb_line]->Display(screen, index_thumb_line*(thumb_height + thumb_border_size), thumb_height);
	index_thumb_line++;

	SDL_PollEvent(&event);

	// exit in ESC or 'q'

	if((event.type == SDL_KEYDOWN) && ((event.key.keysym.sym == 27) || (event.key.keysym.sym == 113))) {
	    goto exit_app;
	} // if

    } // while

 stop_loading:

    image_open = false;
    p_big_image = NULL;

    BackupImage(backup_screen, screen);

    for(;;) {

	while(SDL_PollEvent(&event) == 0) {usleep(10000);}

	switch(event.type) {

	    case SDL_MOUSEMOTION:
		if(image_open) {
		    if((event.motion.x < location.x) || (event.motion.x > (location.x + location.w)) || (event.motion.y < location.y) || (event.motion.y > (location.y + location.h))) {
			RestoreImage(screen, backup_screen, location, &image_open);
		    } // if
		} // if
		break;

	    case SDL_MOUSEBUTTONDOWN:
		if(!image_open) {
		    if(event.button.button == SDL_BUTTON_LEFT) {
			clicked_thumb_line = event.motion.y / (thumb_height+thumb_border_size) + offset_display_index;
			if(clicked_thumb_line >= index_thumb_line) break;
			
			if(p_big_image) {

			    // we probably don't need this anymore

			    p_big_image->UnloadImage();
			} // if

			p_big_image = p_thumb_line_array[clicked_thumb_line]->GetImage(event.motion.x);
			if(!p_big_image) break;

			location.x = p_big_image->GetXStart();
			location.y = (clicked_thumb_line - offset_display_index)*(thumb_height+thumb_border_size);
			location.w = p_big_image->GetImage()->w;
			location.h = p_big_image->GetImage()->h;

			if((location.w < window_width/2) && (location.h < window_height/2)) {
			    location.w *= 2;
			    location.h *= 2;
			} // if
			
			if(location.w > window_width) {
			    location.h = (long)location.h*window_width/(long)location.w;
			    location.w = window_width;
			} // if

			if(location.h > window_height) {
			    location.w = (long)location.w*window_height/(long)location.h;
			    location.h = window_height;
			} // if

			if((location.x + location.w) > window_width) location.x = window_width - location.w;
			if((location.y + location.h) > window_height) location.y = window_height - location.h;

			p_big_image->DisplayImage(screen, &location);
			image_open = true;
		    } // if
		    break;
		} // if
		else {
		    // typ to double size of the image

		    location.w *= 2;
		    location.h *= 2;

		    if(location.w > window_width) {
			location.h = (long)location.h*window_width/(long)location.w;
			location.w = window_width;
		    } // if

		    if(location.h > window_height) {
			location.w = (long)location.w*window_height/(long)location.h;
			location.h = window_height;
		    } // if
		    
		    if((location.x + location.w) > window_width) location.x = window_width - location.w;
		    if((location.y + location.h) > window_height) location.y = window_height - location.h;
		    p_big_image->DisplayImage(screen, &location);
		    image_open = true;
		} // else
		break;

	    case SDL_MOUSEBUTTONUP:
		break;

	    case SDL_KEYDOWN:

		switch(event.key.keysym.sym) {
		    case 27: // ESC
		    case 113: // q
			goto exit_app;
			break;

		    case 274: // Cursor down

			if(!p_thumb_line_array[index_thumb_line]) 
			    p_thumb_line_array[index_thumb_line] = LoadNextThumbLine(p_next_filename, window_width, thumb_quality, thumb_height, thumb_border_size);
			if(p_thumb_line_array[index_thumb_line]) {
			    RestoreImage(screen, backup_screen, location, &image_open);

			    // scroll up

			    offset_display_index++;
			    
			    full_image_rect.x = 0; full_image_rect.y = (thumb_height+thumb_border_size);
			    full_image_rect.w = window_width; full_image_rect.h = window_height - (thumb_height+thumb_border_size);
			    
			    SDL_BlitSurface(screen, &full_image_rect, backup_screen, &full_image_rect);
			    scrolled_rect.x = 0; scrolled_rect.y = 0;
			    scrolled_rect.w = window_width; scrolled_rect.h = window_height - (thumb_height+thumb_border_size);
			    
			    SDL_BlitSurface(backup_screen, &full_image_rect, screen, &scrolled_rect);
			    scrolled_rect.x = 0; scrolled_rect.y = window_height - (thumb_height+thumb_border_size);
			    scrolled_rect.w = window_width; scrolled_rect.h = (thumb_height+thumb_border_size);
			    SDL_FillRect(screen, &scrolled_rect, 0x000000);
			    SDL_UpdateRect(screen, 0, 0, window_width, window_height);
			    
			    // load next thumb line
			    
			    p_thumb_line_array[index_thumb_line]->Display(screen, (index_thumb_line - offset_display_index)*(thumb_height+thumb_border_size), thumb_height);
			    index_thumb_line++;			
			} // if

			BackupImage(backup_screen, screen);
			break;

		    case 273: // Cursor up

			RestoreImage(screen, backup_screen, location, &image_open);
			// scroll down

			if(offset_display_index > 0) {
			    offset_display_index--;
			    index_thumb_line--;
			    
			    full_image_rect.x = 0; full_image_rect.y = 0;
			    full_image_rect.w = window_width; full_image_rect.h = window_height - (thumb_height+thumb_border_size);
			    
			    SDL_BlitSurface(screen, &full_image_rect, backup_screen, &full_image_rect);
			    scrolled_rect.x = 0; scrolled_rect.y = thumb_height+thumb_border_size;
			    scrolled_rect.w = window_width; scrolled_rect.h = window_height - (thumb_height+thumb_border_size);
			    
			    SDL_BlitSurface(backup_screen, &full_image_rect, screen, &scrolled_rect);
			    scrolled_rect.x = 0; scrolled_rect.y = 0;
			    scrolled_rect.w = window_width; scrolled_rect.h = (thumb_height+thumb_border_size);
			    SDL_FillRect(screen, &scrolled_rect, 0x000000);
			    SDL_UpdateRect(screen, 0, 0, window_width, window_height);

			    p_thumb_line_array[index_thumb_line - no_displayable_lines]->Display(screen, 0, thumb_height); 
			} // if

			BackupImage(backup_screen, screen);
			break;

		    default:
			RestoreImage(screen, backup_screen, location, &image_open);
//			cout << event.key.keysym.sym << endl;
			break;
		} // switch
	} // switch
    } // for

 exit_app:
/*
    FILE* p_index = fopen(".opsINDEX", "w");
    FILE* p_cache = fopen(".opsTHUMBS", "w");

    for(int i = 0; i < MAX_THUMB_LINES; i++)
	if(p_thumb_line_array[i]) p_thumb_line_array[i]->Save(p_index, p_cache);

    fclose(p_index);
    fclose(p_cache);
*/
    for(int i = 0; i < index_thumb_line; i++) {
	delete p_thumb_line_array[i];
    } // for
   
    delete p_filenames; 

    SDL_FreeSurface(screen);
    SDL_FreeSurface(backup_screen);
} // main
